package pl.edu.icm.yadda.service2.user;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import pl.edu.icm.yadda.aas.usercatalog.model.User;
import pl.edu.icm.yadda.service2.user.credential.Credential;
import pl.edu.icm.yadda.service2.user.credential.LoginPasswordCredential;
import pl.edu.icm.yadda.service2.user.credential.Credential.STATUS;
import pl.edu.icm.yadda.service2.user.model.Group;
import pl.edu.icm.yadda.service2.user.model.GroupName;
import pl.edu.icm.yadda.service2.user.model.UserAttributesConstants;
import pl.edu.icm.yadda.service2.user.model.UserData;
import pl.edu.icm.yadda.service2.user.model.UserFlagsConstants;

/**
 * User data converter class.
 * @author mhorst
 *
 */
public class UserDataConverter {

	
	/**
	 * Converts {@link User} from user catalog model into {@link UserData}.
	 * @param source
	 * @param predefinedDomain
	 * @param setNameAsInternalId flag indicating source user name should be taken either as internal id 
	 * or or as external id causing internal id to be generated
	 * @return converted {@link UserData} object
	 * @throws UserDataConverterException
	 */
	public static UserData convert(User source, 
			String predefinedDomain, boolean setNameAsInternalId) throws UserDataConverterException {
		if (source==null) {
			return null;
		}
		if (source.isDeleted()) {
//			this seems to be the most suitable approach, disallowing conversion of deleted user
			throw new UserDataConverterException("unable to convert user object: " + 
					source.getName() + ", reason: object flagged as deleted!");
		}
		if (source.getName()!=null) {
			UserData userDataResult = new UserData();
			userDataResult.setUser(new pl.edu.icm.yadda.service2.user.model.User());
//			using name/login as unique id
			if (setNameAsInternalId) {
				userDataResult.getUser().setId(source.getName());	
			} else {
				userDataResult.getUser().setIdentifiers(new HashSet<String>());
				userDataResult.getUser().getIdentifiers().add(source.getName());
			}
			userDataResult.getUser().setDomain(predefinedDomain);
			userDataResult.getUser().setFlags(new HashSet<String>());
			if (source.isActivated()) {
				userDataResult.getUser().getFlags().add(UserFlagsConstants.ACTIVE);
			} else {
				userDataResult.getUser().getFlags().add(UserFlagsConstants.INACTIVE);
			}
			userDataResult.getUser().setAttributes(source.getAttributes());
//			setting email as additional attribute
			if (source.getEmail()!=null) {
				if (userDataResult.getUser().getAttributes()==null) {
					userDataResult.getUser().setAttributes(new HashMap<String, String>());
				}
				userDataResult.getUser().getAttributes().put(
						UserAttributesConstants.EMAIL, source.getEmail());
			}
			userDataResult.getUser().setRoles(source.getRoles());
//			setting direct groups, will be handled by UserDBInitializer
			userDataResult.setDirectGroups(
					prepareGroups(source.getGroups(), predefinedDomain));
//			setting credential, will be handled by UserDBInitializer
			if (source.getPassword()!=null && 
					source.getPassword().trim().length()>0) {
				userDataResult.setCredentials(
						prepareCredentials(source.getPassword()));	
			}
//			licenses were not used in old model, omitting
			if (source.getLicenses()!=null && !source.getLicenses().isEmpty()) {
				System.err.println("licenses were defined for user " + source.getName() +
						" but won't be propagated to userdb!");
			}
//			those two will not be part of userdb anymore
//			source.getProfile();
//			source.getPreferences();
			return userDataResult;
			
		} else {
			throw new UserDataConverterException("source user name was not defined!");
		}
	}
	
	protected static Set<GroupName> prepareGroups(Set<String> groups, String predefinedDomain) {
		if (groups!=null) {
			if (!groups.isEmpty()) {
				Set<GroupName> result = new HashSet<GroupName>(groups.size());
				for (String group : groups) {
					result.add(new GroupName(predefinedDomain, group));
				}
				return result;
			} else {
				return Collections.emptySet();
			}
		} else {
			return null;
		}
	}
	
	protected static List<Credential> prepareCredentials(String password) {
		LoginPasswordCredential cred = new LoginPasswordCredential();
		cred.setPassword(password);
		cred.setStatus(STATUS.ACTIVE);
		List<Credential> result = new ArrayList<Credential>(1);
		result.add(cred);
		return result;
	}
	
	/**
	 * Converts {@link pl.edu.icm.yadda.aas.usercatalog.model.Group} from user catalog model into {@link Group}.
	 * @param source
	 * @param predefinedDomain
	 * @return converted {@link Group} object
	 * @throws UserDataConverterException
	 */
	public static Group convert(pl.edu.icm.yadda.aas.usercatalog.model.Group source,
			String predefinedDomain) throws UserDataConverterException {
		if (source==null) {
			return null;
		}
		if (source.getName()!=null) {
			Group result = new Group();
			result.setGroupName(new GroupName(predefinedDomain, source.getName()));
			result.setRoles(source.getRoles());
//			parent group will be set at postprocessing phase
			if (source.getAddresses()!=null && source.getAddresses().size()>0) {
				System.err.println("source addresses for group " + source.getName() + 
						" will not be transferred to new model!");
			}
			if (source.getAttributes()!=null && source.getAttributes().size()>0) {
				System.err.println("source attributes for group " + source.getName() + 
						" will not be transferred to new model!");
			}
			if (source.getLicenses()!=null && source.getLicenses().size()>0) {
				System.err.println("source lisenses for group " + source.getName() + 
						" will not be transferred to new model!");
			}
			if (source.getOwners()!=null && source.getOwners().size()>0) {
				System.err.println("source owners for group " + source.getName() + 
						" will not be transferred to new model!");
			}
			return result;	
		} else {
			throw new UserDataConverterException("source group name was not defined!");
		}
	}
	
}
