package pl.edu.icm.yadda.service2.user;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.springframework.context.ApplicationContext;

import pl.edu.icm.yadda.service2.catalog.CountingIterator;
import pl.edu.icm.yadda.service2.user.UserDataMigrator.InvalidOptionException;
import pl.edu.icm.yadda.service2.user.exception.ExportException;
import pl.edu.icm.yadda.spring.utils.SpringUtils;

/**
 * User data exporter module.
 * Exports groups and users to intermediate format.
 * @author mhorst
 *
 */
public class UserDataExporter {

	public static final String PARAM_SOURCE_DB_ADDR = "sdb_addr";
	public static final String PARAM_SOURCE_DB_LOGIN = "sdb_login";
	public static final String PARAM_SOURCE_DB_PASS = "sdb_pass";
	public static final String PARAM_PREDEFINED_DOMAIN = "d";
	public static final String PARAM_TARGET_DIR = "dir";
	public static final String PARAM_HELP = "h";
	
	public static final String PROPERTY_SOURCE_DB_ADDR 	= "usercatalog.db.url";
	public static final String PROPERTY_SOURCE_DB_LOGIN = "usercatalog.db.username";
	public static final String PROPERTY_SOURCE_DB_PASS 	= "usercatalog.db.password";
//	not needed but placeholder in context has to be set
	public static final String PROPERTY_OVERWRITE_EXISTING 	= "user.data.migrator.overwrite.existing";
	
	protected static String defaultEncoding = "UTF-8";
	
	public static void main(String[] args) {
		
		Options options = new Options();
		options.addOption(PARAM_SOURCE_DB_ADDR, true, "source database address");
		options.addOption(PARAM_SOURCE_DB_LOGIN, true, "source database login");
		options.addOption(PARAM_SOURCE_DB_PASS, true, "source database password");
		options.addOption(PARAM_PREDEFINED_DOMAIN, true, "predefined domain to be set, required!");
		options.addOption(PARAM_TARGET_DIR, true, "target directory " +
				"where all data should be exported, curent by default");
		options.addOption(PARAM_HELP, false, "shows help");
		
		try {
			CommandLineParser parser = new PosixParser();
			CommandLine cl = null;
			try {			
				cl = parser.parse(options, args);
				if (cl.hasOption(PARAM_HELP)) {
					HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
					return;
				}
//				setting system properties utilized by spring property placeholder
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_SOURCE_DB_ADDR, PROPERTY_SOURCE_DB_ADDR);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_SOURCE_DB_LOGIN, PROPERTY_SOURCE_DB_LOGIN);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_SOURCE_DB_PASS, PROPERTY_SOURCE_DB_PASS);
				System.setProperty(PROPERTY_OVERWRITE_EXISTING, "false");
				
				String domain = null; 
				if (cl.hasOption(PARAM_PREDEFINED_DOMAIN)) {
					domain = cl.getOptionValue(PARAM_PREDEFINED_DOMAIN);
				} else {
					System.err.println("required parameter is missing: " + PARAM_PREDEFINED_DOMAIN);
					HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
					System.exit(1);
				}
				
				String dirStr = null;
				if (cl.hasOption(PARAM_TARGET_DIR)) {
					dirStr = cl.getOptionValue(PARAM_TARGET_DIR);
				} else {
					dirStr = System.getProperty("user.dir");
				}
				dirStr = dirStr + File.separatorChar + "export_" + System.currentTimeMillis();
				
//				creating directory
				File dir = new File(dirStr);
				dir.mkdirs();
				
				ApplicationContext appCtx = SpringUtils.getSpringContext(
				"classpath:/pl/edu/icm/yadda/service2/user/user-data-exporter-context.xml");
				UserCatalog userDb = (UserCatalog) appCtx.getBean("userCatalog");
				{
//					exporting users
					CountingIterator<String> usersIt = userDb.iterateUsers(new String[] {domain});
					System.out.println("exporting " + usersIt.count() + " users to directory: " + dirStr);
					while(usersIt.hasNext()) {
						OutputStreamWriter writer = null;
						try {
							String userId = usersIt.next();
							System.out.println("exporting user " + userId);
							File targetFile = new File(dir, UserCatalogConstants.USER_PREFIX
									+ userId + ".xml");
							writer = new OutputStreamWriter(new FileOutputStream(targetFile), defaultEncoding);
							writer.append(userDb.exportEntity(userId, UserCatalogConstants.USER_XML));
						} finally {
							if (writer!=null) {
								writer.close();
							}
						}
					}
				}
				{
//					exporting groups
					CountingIterator<String> groupsIt = userDb.iterateGroups(new String[] {domain});
					System.out.println("exporting " + groupsIt.count() + " groups to directory: " + dirStr);
					while(groupsIt.hasNext()) {
						OutputStreamWriter writer = null;
						try {
							String groupId = groupsIt.next();
							System.out.println("exporting group " + groupId);
							File targetFile = new File(dir, UserCatalogConstants.GROUP_PREFIX + 
									groupId + ".xml");
							writer = new OutputStreamWriter(new FileOutputStream(targetFile), defaultEncoding);
							writer.append(userDb.exportEntity(groupId, UserCatalogConstants.GROUP_XML));
						} finally {
							if (writer!=null) {
								writer.close();
							}
						}
					}
				}
				
			} catch (ParseException e) {
				throw new InvalidOptionException(e.getMessage());
			} catch (ExportException e) {
				throw new RuntimeException("unexpected exception while exporting user", e);
			} catch (IOException e) {
				throw new RuntimeException("unexpected exception while writing user to file", e);
			}
		} catch (InvalidOptionException e) {
			if (e.getMessage() != null)	System.err.println("Error: "+e.getMessage());
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
			System.exit(1);
		}	
	}
}
