package pl.edu.icm.yadda.service2.user;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.springframework.context.ApplicationContext;

import pl.edu.icm.yadda.service2.user.UserDataMigrator.InvalidOptionException;
import pl.edu.icm.yadda.service2.user.domain.DomainOverrider;
import pl.edu.icm.yadda.service2.user.exception.ImportException;
import pl.edu.icm.yadda.spring.utils.SpringUtils;

/**
 * User data importer module.
 * Imports groups and users from intermediate format.
 * @author mhorst
 *
 */
public class UserDataImporter {

	public static final String PARAM_TARGET_DB_ADDR = "tdb_addr";
	public static final String PARAM_TARGET_DB_LOGIN = "tdb_login";
	public static final String PARAM_TARGET_DB_PASS = "tdb_pass";
	public static final String PARAM_PREDEFINED_DOMAIN = "d";
	public static final String PARAM_SOURCE_DIR = "dir";
	public static final String PARAM_HELP = "h";
	
	public static final String PROPERTY_TARGET_DB_ADDR 	= "usercatalog.db.url";
	public static final String PROPERTY_TARGET_DB_LOGIN = "usercatalog.db.username";
	public static final String PROPERTY_TARGET_DB_PASS 	= "usercatalog.db.password";

	public static final String PROPERTY_OVERWRITE_EXISTING 	= "user.data.migrator.overwrite.existing";
	
	protected static String defaultEncoding = "UTF-8";
	
	public static void main(String[] args) {
		
		Options options = new Options();
		options.addOption(PARAM_TARGET_DB_ADDR, true, "target database address");
		options.addOption(PARAM_TARGET_DB_LOGIN, true, "target database login");
		options.addOption(PARAM_TARGET_DB_PASS, true, "target database password");
		options.addOption(PARAM_PREDEFINED_DOMAIN, true, "target domain, if not set domain will not be overriden");
		options.addOption(PARAM_SOURCE_DIR, true, "source directory " +
				"where all data should be imported from, current by default");
		options.addOption(PARAM_HELP, false, "shows help");
		
		try {
			CommandLineParser parser = new PosixParser();
			CommandLine cl = null;
			try {			
				cl = parser.parse(options, args);
				if (cl.hasOption(PARAM_HELP)) {
					HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
					return;
				}
//				setting system properties utilized by spring property placeholder
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_TARGET_DB_ADDR, PROPERTY_TARGET_DB_ADDR);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_TARGET_DB_LOGIN, PROPERTY_TARGET_DB_LOGIN);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_TARGET_DB_PASS, PROPERTY_TARGET_DB_PASS);
				
				System.setProperty(PROPERTY_OVERWRITE_EXISTING, "false");
				
				String targetDomain = cl.getOptionValue(PARAM_PREDEFINED_DOMAIN); 
				if (targetDomain!=null) {
					System.out.println("overwritting target domain to: " + targetDomain);
				} else {
					System.out.println("no target domain set, using domains defined within import source files");
				}
				
				String dirStr = null;
				if (cl.hasOption(PARAM_SOURCE_DIR)) {
					dirStr = cl.getOptionValue(PARAM_SOURCE_DIR);
				} else {
					dirStr = System.getProperty("user.dir");
				}
				File dirFile = new File(dirStr);
				
				ApplicationContext appCtx = SpringUtils.getSpringContext(
				"classpath:/pl/edu/icm/yadda/service2/user/user-data-importer-context.xml");
				UserCatalog userDb = (UserCatalog) appCtx.getBean("userCatalog");
				DomainOverrider userDomainOverrider = (DomainOverrider) appCtx.getBean("userDomainOverrider");
				DomainOverrider groupDomainOverrider = (DomainOverrider) appCtx.getBean("groupDomainOverrider");
				
				{
//					importing groups
					File[] groupFiles = dirFile.listFiles(new FilenameFilter() {
						@Override
						public boolean accept(File dir, String name) {
							return name.toUpperCase().startsWith(UserCatalogConstants.GROUP_PREFIX);
						}
					});
					if (groupFiles!=null && groupFiles.length>0) {
						System.out.println("importing " + groupFiles.length + " groups");
						for (File groupFile : groupFiles) {
							BufferedReader reader = null;
							StringBuffer strBuff = new StringBuffer();
							try {
								reader = new BufferedReader(new InputStreamReader(
										new FileInputStream(groupFile), defaultEncoding));
								String line;
								while((line = reader.readLine())!=null) {
									strBuff.append(line);
								}
								userDb.importEntity(null, UserCatalogConstants.GROUP_XML, 
										targetDomain!=null?
												groupDomainOverrider.overrideDomain(strBuff.toString(), targetDomain):
													strBuff.toString());
							} finally {
								if (reader!=null) {
									reader.close();
								}
							}
						}
					} else {
						System.out.println("no groups to be imported");
					}
				}
				{
//					importing users
					File[] userFiles = dirFile.listFiles(new FilenameFilter() {
						@Override
						public boolean accept(File dir, String name) {
							return name.toUpperCase().startsWith(UserCatalogConstants.USER_PREFIX);
						}
					});
					if (userFiles!=null && userFiles.length>0) {
						System.out.println("importing " + userFiles.length + " users");
						for (File userFile : userFiles) {
							BufferedReader reader = null;
							StringBuffer strBuff = new StringBuffer();
							try {
								reader = new BufferedReader(new InputStreamReader(
										new FileInputStream(userFile), defaultEncoding));
								String line;
								while((line = reader.readLine())!=null) {
									strBuff.append(line);
								}
								userDb.importEntity(null, UserCatalogConstants.USER_XML, 
								targetDomain!=null?
										userDomainOverrider.overrideDomain(strBuff.toString(), targetDomain):
											strBuff.toString());
							} finally {
								if (reader!=null) {
									reader.close();
								}
							}
						}
					} else {
						System.out.println("no users to be imported");
					}
				}
				
			} catch (ParseException e) {
				throw new InvalidOptionException(e.getMessage());
			} catch (ImportException e) {
				throw new RuntimeException("unexpected exception while importing user", e);
			} catch (IOException e) {
				throw new RuntimeException("unexpected exception while reading user from file", e);
			}
		} catch (InvalidOptionException e) {
			if (e.getMessage() != null)	System.err.println("Error: "+e.getMessage());
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
			System.exit(1);
		}	
	}
}
