package pl.edu.icm.yadda.service2.user;

import java.io.Serializable;
import java.util.Iterator;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.springframework.context.ApplicationContext;

import pl.edu.icm.yadda.aal.service2.IUserManagerServiceFacade;
import pl.edu.icm.yadda.aas.usercatalog.model.Group;
import pl.edu.icm.yadda.aas.usercatalog.model.User;
import pl.edu.icm.yadda.aas.usercatalog.service.BrowseSecurityObjectResponse;
import pl.edu.icm.yadda.aas.usercatalog.service.BrowseSecurityObjectsRequest;
import pl.edu.icm.yadda.aas.usercatalog.service.IUserCatalogService;
import pl.edu.icm.yadda.aas.usercatalog.service.SecurityObjectType;
import pl.edu.icm.yadda.aas.userdb.UserDBInitializer;
import pl.edu.icm.yadda.service2.browse.FetchRequest;
import pl.edu.icm.yadda.service2.browse.FetchRequest.Page;
import pl.edu.icm.yadda.service2.exception.ServiceException;
import pl.edu.icm.yadda.service2.user.log.MigrationLogger;
import pl.edu.icm.yadda.service2.user.log.MigrationLogger.Status;
import pl.edu.icm.yadda.service2.user.log.MigrationLogger.Type;
import pl.edu.icm.yadda.service2.user.model.GroupName;
import pl.edu.icm.yadda.spring.utils.SpringUtils;

/**
 * User data migrator importing data retrieved from
 * {@link IUserCatalogService} into new {@link UserCatalog}.
 * 
 * @author mhorst
 *
 */
public class UserDataMigrator {

	public static final String PARAM_SOURCE_DB_ADDR = "sdb_addr";
	public static final String PARAM_SOURCE_DB_LOGIN = "sdb_login";
	public static final String PARAM_SOURCE_DB_PASS = "sdb_pass";
	
	public static final String PARAM_TARGET_DB_ADDR = "tdb_addr";
	public static final String PARAM_TARGET_DB_LOGIN = "tdb_login";
	public static final String PARAM_TARGET_DB_PASS = "tdb_pass";
	
	public static final String PARAM_PREDEFINED_DOMAIN = "d";
	public static final String PARAM_OVERWRITE_EXISTING = "o";
	public static final String PARAM_PAGE_SIZE = "page";
	public static final String PARAM_HELP = "h";
	public static final String PARAM_INTERNAL_ID = "intId";
	public static final String PARAM_AAL3_MODE = "aal3";
	
	public static final String PROPERTY_SOURCE_DB_ADDR 	= "catalog.main.db.psql.url_catalog";
	public static final String PROPERTY_SOURCE_DB_LOGIN = "catalog.main.db.psql.username";
	public static final String PROPERTY_SOURCE_DB_PASS 	= "catalog.main.db.psql.password";
	
	public static final String PROPERTY_TARGET_DB_ADDR 	= "usercatalog.db.url";
	public static final String PROPERTY_TARGET_DB_LOGIN = "usercatalog.db.username";
	public static final String PROPERTY_TARGET_DB_PASS 	= "usercatalog.db.password";
	
	public static final String PROPERTY_OVERWRITE_EXISTING 	= "user.data.migrator.overwrite.existing";
	
	public static final int DEFAULT_USER_PAGE_SIZE = 50;
	public static final boolean DEFAULT_OVERWRITE_EXISTING = false;

	
	@SuppressWarnings("serial")
	static class InvalidOptionException extends Exception {

		/**
		 * @param msg the error message
		 */
		public InvalidOptionException(String msg) {
			super(msg);
		}

		/**
		 * Creates an exception without any error message
		 */
		public InvalidOptionException() {}		
	}
	
	public static void main(String[] args) {
	
		Options options = new Options();
		options.addOption(PARAM_SOURCE_DB_ADDR, true, "source database address");
		options.addOption(PARAM_SOURCE_DB_LOGIN, true, "source database login");
		options.addOption(PARAM_SOURCE_DB_PASS, true, "source database password");
		
		options.addOption(PARAM_TARGET_DB_ADDR, true, "target database address");
		options.addOption(PARAM_TARGET_DB_LOGIN, true, "target database login");
		options.addOption(PARAM_TARGET_DB_PASS, true, "target database password");
		
		options.addOption(PARAM_PREDEFINED_DOMAIN, true, "predefined domain to be set, required!");
		options.addOption(PARAM_PAGE_SIZE, true, "maximum number of users being processed for single query, set to 50 by default");
		options.addOption(PARAM_OVERWRITE_EXISTING, false, "overwrites existing users, skips existing when not set");
		options.addOption(PARAM_INTERNAL_ID, false, "flag indicating source user name should be taken as internal id. " +
				"By default will be set as external id causing internal id to be generated");
		options.addOption(PARAM_AAL3_MODE, false, "sets migrator into AAL3 mode, by default AAL2 mode is handled when reading source data");
		options.addOption(PARAM_HELP, false, "shows help");
		
		try {
			CommandLineParser parser = new PosixParser();
			CommandLine cl = null;
			try {			
				cl = parser.parse(options, args);
				if (cl.hasOption(PARAM_HELP)) {
					HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
					return;
				}
//				setting system properties utilized by spring property placeholder
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_SOURCE_DB_ADDR, PROPERTY_SOURCE_DB_ADDR);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_SOURCE_DB_LOGIN, PROPERTY_SOURCE_DB_LOGIN);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_SOURCE_DB_PASS, PROPERTY_SOURCE_DB_PASS);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_TARGET_DB_ADDR, PROPERTY_TARGET_DB_ADDR);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_TARGET_DB_LOGIN, PROPERTY_TARGET_DB_LOGIN);
				UserDataMigratorHelper.setSystemProperty(options, cl, 
						PARAM_TARGET_DB_PASS, PROPERTY_TARGET_DB_PASS);
				System.setProperty(PROPERTY_OVERWRITE_EXISTING, 
						Boolean.toString(cl.hasOption(PARAM_OVERWRITE_EXISTING)?
								true:DEFAULT_OVERWRITE_EXISTING));
				int pageSize = cl.hasOption(PARAM_PAGE_SIZE)?
						Integer.parseInt(cl.getOptionValue(PARAM_PAGE_SIZE)):DEFAULT_USER_PAGE_SIZE;
				String domain = null; 
				if (cl.hasOption(PARAM_PREDEFINED_DOMAIN)) {
					domain = cl.getOptionValue(PARAM_PREDEFINED_DOMAIN);
					if (!domain.endsWith(UserCatalogConstants.DL_DOMAIN_SUFFIX)) {
//						supplementing with DL suffix
						System.out.println("supplementing received domain '" + 
								domain +"' with required suffix: '" + 
								UserCatalogConstants.DL_DOMAIN_SUFFIX + "'!");
						domain = domain + UserCatalogConstants.DL_DOMAIN_SUFFIX;
					}
				} else {
					System.err.println("required parameter is missing: " + PARAM_PREDEFINED_DOMAIN);
					HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
					System.exit(1);
				}
				boolean setNameAsInternalId = cl.hasOption(PARAM_INTERNAL_ID);
				
				if (cl.hasOption(PARAM_AAL3_MODE)) {
					System.out.println("working in AAL3 mode");
					ApplicationContext appCtx = SpringUtils.getSpringContext(
						"classpath:/pl/edu/icm/yadda/service2/user/user-data-migrator-context-aal3.xml");
					IUserCatalogService sourceUserCatalog = (IUserCatalogService) appCtx.getBean("localUserCatalogService");
					UserCatalog targetUserDb = (UserCatalog) appCtx.getBean("userCatalog");
					UserDBInitializer userDBInitializer = (UserDBInitializer) appCtx.getBean("userDBInitializer");
					MigrationLogger logger = (MigrationLogger) appCtx.getBean("migrationLogger");
					try {
						logger.init();
						{
		//					migrating groups
							BrowseSecurityObjectsRequest browseReq = new BrowseSecurityObjectsRequest(
									SecurityObjectType.GROUP, null);
							browseReq.setLimit(pageSize);
							BrowseSecurityObjectResponse browseResp = sourceUserCatalog.browseObjects(browseReq);
							int counter = 0;
							while(browseResp.getResults()!=null && browseResp.getResults().size()>0) {
								for (Serializable result : browseResp.getResults()) {
									if (result instanceof Group) {
										Group sourceGroup = (Group) result;
										try {
											boolean written = userDBInitializer.handleStoringSingleGroup(
													UserDataConverter.convert(sourceGroup, domain));
											logger.log(Type.GROUP, written?Status.OK:Status.SKIPPED, 
													sourceGroup.getName(), "");
											counter++;
										} catch (Exception e) {
											System.err.println("Error occurred when processing group: " + 
													sourceGroup.getName());
											e.printStackTrace(System.err);
											logger.log(Type.GROUP, Status.FAIL, 
													sourceGroup.getName(), e.getMessage());
											System.exit(1);
										}
									} else {
										System.err.println("unexpected object instance! " +
												"expected: " + Group.class + " got: " + result.getClass());
										System.exit(1);
									}
								}
		//						getting next page of data when available
								if (browseResp.getCookie()!=null) {
									browseResp = sourceUserCatalog.fetchMoreObjects(new FetchRequest(
											browseResp.getCookie(), Page.NEXT, pageSize));
								} else {
									break;
								}
							}
		//					second iteration: resolving relationship after storing all groups (group existance required)
							browseReq = new BrowseSecurityObjectsRequest(
									SecurityObjectType.GROUP, null);
							browseReq.setLimit(pageSize);
							browseResp = sourceUserCatalog.browseObjects(browseReq);
							while(browseResp.getResults()!=null && browseResp.getResults().size()>0) {
								for (Serializable result : browseResp.getResults()) {
									Group sourceGroup = (Group) result;
									try {
										if (sourceGroup.getSupergroup()!=null) {
											System.out.println("setting: " + sourceGroup.getSupergroup() + 
													" as parent group for group: " + sourceGroup.getName());
											targetUserDb.assignGroup(
													new GroupName(domain, sourceGroup.getName()), 
													new GroupName(domain, sourceGroup.getSupergroup()));
										}
									} catch (Exception e) {
										System.err.println("Error occurred when processing group: " + 
												sourceGroup.getName());
										e.printStackTrace(System.err);
										System.exit(1);
									}
								}
		//						getting next page of data when available
								if (browseResp.getCookie()!=null) {
									browseResp = sourceUserCatalog.fetchMoreObjects(new FetchRequest(
											browseResp.getCookie(), Page.NEXT, pageSize));
								} else {
									break;
								}
							}
							
							System.out.println("number of group objects processed: " + counter);
		//					checking whether no error was returned
							if (!browseResp.isOK()) {
								System.err.println("got error when paging results: " + 
										browseResp.getError().getMssg());
								browseResp.getError().getException().printStackTrace(System.err);
								System.exit(1);
							}
						}
						
						{
		//					migrating users
							BrowseSecurityObjectsRequest browseReq = new BrowseSecurityObjectsRequest(
									SecurityObjectType.USER, null);
							browseReq.setLoadPasswords(true);
							browseReq.setLimit(pageSize);
							BrowseSecurityObjectResponse browseResp = sourceUserCatalog.browseObjects(browseReq);
							int counter = 0;
							while(browseResp.getResults()!=null && browseResp.getResults().size()>0) {
								for (Serializable result : browseResp.getResults()) {
									if (result instanceof User) {
										User sourceUser = (User) result;
										try {
											if (sourceUser.isDeleted()) {
												System.out.println("omitting conversion of deleted user: " + 
														sourceUser.getName());
												logger.log(Type.USER, Status.SKIPPED, 
														sourceUser.getName(), "deleted flag set in source object");
											} else {
												boolean written = userDBInitializer.handleStoringSingleUser(
														UserDataConverter.convert(sourceUser, 
																domain, setNameAsInternalId));
												logger.log(Type.USER, written?Status.OK:Status.SKIPPED, 
														sourceUser.getName(), "");
											}
											counter++;
										} catch (Exception e) {
											System.err.println("Error occurred when processing user: " + 
													sourceUser.getName());
											e.printStackTrace(System.err);
											logger.log(Type.USER, Status.FAIL, 
													sourceUser.getName(), e.getMessage());
											System.exit(1);
										}
									} else {
										System.err.println("unexpected object instance! " +
												"expected: " + User.class + " got: " + result.getClass());
										System.exit(1);
									}
								}
			//					getting next page of data when available
								if (browseResp.getCookie()!=null) {
									browseResp = sourceUserCatalog.fetchMoreObjects(new FetchRequest(
											browseResp.getCookie(), Page.NEXT, pageSize));
								} else {
									break;
								}
							}
							System.out.println("number of user objects processed: " + counter);
			//				checking whether no error was returned
							if (!browseResp.isOK()) {
								System.err.println("got error when paging results: " + 
										browseResp.getError().getMssg());
								browseResp.getError().getException().printStackTrace(System.err);
								System.exit(1);
							}
						}
		
						
					} finally {
						logger.close();
					}

				} else {
					System.out.println("working in AAL2 mode");
					ApplicationContext appCtx = SpringUtils.getSpringContext(
						"classpath:/pl/edu/icm/yadda/service2/user/user-data-migrator-context-aal2.xml");
					IUserManagerServiceFacade sourceUserCatalogFacade = (IUserManagerServiceFacade) appCtx.getBean("userManagerServiceFacade");
					
					UserCatalog targetUserDb = (UserCatalog) appCtx.getBean("userCatalog");
					UserDBInitializer userDBInitializer = (UserDBInitializer) appCtx.getBean("userDBInitializer");
					MigrationLogger logger = (MigrationLogger) appCtx.getBean("migrationLogger");
					try {
						logger.init();
						{
		//					migrating groups
							int counter = 0;
							try {
								Iterator<pl.edu.icm.yadda.aal.model2.Group> groupIt = sourceUserCatalogFacade.iterateGroups();
								while(groupIt.hasNext()) {
									pl.edu.icm.yadda.aal.model2.Group sourceGroup = groupIt.next();
									try {
										boolean written = userDBInitializer.handleStoringSingleGroup(
												AAL2UserDataConverter.convert(sourceGroup, domain));
										logger.log(Type.GROUP, written?Status.OK:Status.SKIPPED, 
												sourceGroup.getName(), "");
										counter++;
									} catch (Exception e) {
										System.err.println("Error occurred when processing group: " + 
												sourceGroup.getName());
										e.printStackTrace(System.err);
										logger.log(Type.GROUP, Status.FAIL, 
												sourceGroup.getName(), e.getMessage());
										System.exit(1);
									}
								}
			//					second iteration: resolving relationship after storing all groups (group existance required)
								groupIt = sourceUserCatalogFacade.iterateGroups();
								while(groupIt.hasNext()) {
									pl.edu.icm.yadda.aal.model2.Group sourceGroup = groupIt.next();
									try {
										if (sourceGroup.getSupergroup()!=null) {
											if (!sourceGroup.getSupergroup().isResolved()) {
												sourceGroup.getSupergroup().resolve();
											} 
											String superGroupName = sourceGroup.getSupergroup().getReference().getName();
											System.out.println("setting: " + superGroupName + 
													" as parent group for group: " + sourceGroup.getName());
											targetUserDb.assignGroup(
													new GroupName(domain, sourceGroup.getName()), 
													new GroupName(domain, superGroupName));
										}
									} catch (Exception e) {
										System.err.println("Error occurred when processing group: " + 
												sourceGroup.getName());
										e.printStackTrace(System.err);
										System.exit(1);
									}
								}
							} catch (ServiceException e) {
								System.err.println("Error occurred when listing groups!");
								e.printStackTrace(System.err);
								System.exit(1);
							} finally {
								System.out.println("number of group objects processed: " + counter);	
							}
						}
						
						{
		//					migrating users
							int counter = 0;
							try {
								Iterator<String> userIdIt = sourceUserCatalogFacade.iterateUserIds(null);
								while(userIdIt.hasNext()) {
									String userId = userIdIt.next();
									try {
										pl.edu.icm.yadda.aal.model2.User sourceUser = sourceUserCatalogFacade.loadUser(userId);
										try {
											if (sourceUser.isDeleted()) {
												System.out.println("omitting conversion of deleted user: " + 
														sourceUser.getLogin());
												logger.log(Type.USER, Status.SKIPPED, 
														sourceUser.getLogin(), "deleted flag set in source object");
											} else {
												boolean written = userDBInitializer.handleStoringSingleUser(
														AAL2UserDataConverter.convert(sourceUser, 
																domain, setNameAsInternalId));
												logger.log(Type.USER, written?Status.OK:Status.SKIPPED, 
														sourceUser.getLogin(), "");
											}
											counter++;
										} catch (Exception e) {
											System.err.println("Error occurred when processing user: " + 
													sourceUser.getLogin());
											e.printStackTrace(System.err);
											logger.log(Type.USER, Status.FAIL, 
													sourceUser.getLogin(), e.getMessage());
											System.exit(1);
										}
									} catch (ServiceException e) {
										System.err.println("Error occurred when processing user: " + 
												userId);
										e.printStackTrace(System.err);
										logger.log(Type.USER, Status.FAIL, 
												userId, e.getMessage());
										System.exit(1);
									}
								}
							} catch (ServiceException e) {
								System.err.println("Error occurred when listing users!");
								e.printStackTrace(System.err);
								System.exit(1);
							} finally {
								System.out.println("number of user objects processed: " + counter);
							}
						}
					} finally {
						logger.close();
					} 
				}
			} catch (ParseException e) {
				throw new InvalidOptionException(e.getMessage());
			}
		} catch (InvalidOptionException e) {
			if (e.getMessage() != null)	System.err.println("Error: "+e.getMessage());
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("java " + UserDataMigrator.class.getCanonicalName(), options);
			System.exit(1);
		}	
	}
	
}
